Imports System.IO
Imports System.Threading
Imports System.ComponentModel
Imports System.Collections.Specialized
Imports System.Reflection

'
' This is the form used for polling the network and performing some
' operation on each PC.  It uses a thread poll to simultaneously poll a 
' number of PCs at once.
'
Public Class fmMain

    Private TotalCount, SuccessCount As Integer
    Private maxThreads, portThreads, PollTimeOut, ListCount As Integer
    Private PingValidation As Boolean
    Private sWatch As New Stopwatch
    Private tw As TextWriter
    Private fs As FileStream
    Private PollLogLevel As PollLogLevels

    Private Enum PollLogLevels As Integer
        None = 0
        Summary
        Errors
        Full
    End Enum

    Private Sub Form1_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        ThreadPool.GetMaxThreads(maxThreads, portThreads)

        nudPoolSize.Maximum = maxThreads
        nudPoolSize.Minimum = 1
        cmbLogLevel.SelectedItem = "Full"
        tbLogfile.Text = Path.Combine(Environment.ExpandEnvironmentVariables(My.Settings.LogDir), "BulkCopy_" & Now.ToString("yyyyMMdd") & ".txt")
    End Sub

    Private Sub bnClose_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles bnClose.Click
        BackgroundWorker1.CancelAsync()
        Thread.Sleep(500)
        Me.Close()
    End Sub

    '
    ' Setup the polling mechanism
    '
    Private Sub bnStart_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles bnStart.Click
        Dim PC_List() As String

        ' a quick sanity check
        If tbSource.Text = "" Or tbDestination.Text = "" Then
            MsgBox("Yikes, You didn't provide a Source or Destination", MsgBoxStyle.Exclamation)
            Exit Sub
        End If

        If Directory.GetFiles(Path.GetDirectoryName(tbSource.Text), Path.GetFileName(tbSource.Text), SearchOption.TopDirectoryOnly).Length = 0 Then
            MsgBox("Yikes, Can't find the source file", MsgBoxStyle.Exclamation)
            Exit Sub
        End If

        ' Tweak the user's input...
        If Not tbDestination.Text.EndsWith("\") Then
            tbDestination.Text &= "\"
        End If

        pbStarted.Value = 0
        pbCompleted.Value = 0
        sWatch.Reset()
        PollLogLevel = CType(cmbLogLevel.SelectedIndex, PollLogLevels)
        PollTimeOut = CInt(nudPollTimeout.Value)
        PingValidation = cbPing.Checked

        ' open the logfile
        tw = Nothing
        fs = Nothing
        PC_List = Nothing
        If tbLogfile.Text <> "" And PollLogLevel > PollLogLevels.None Then
            Try
                fs = New FileStream(tbLogfile.Text, FileMode.Append)
                ' use a thread-safe stream writer
                tw = TextWriter.Synchronized(New StreamWriter(fs))

                tw.WriteLine("Started: " & Now)
                tw.WriteLine("File: " & tbSource.Text)
            Catch ex As Exception
                MsgBox("Can't open the log file" & vbCr & ex.Message, MsgBoxStyle.Exclamation, "Error")
                Exit Sub
            End Try
        End If

        ' Get the list of PCs
        tbStatus.Text = "Getting network list..."
        Application.DoEvents()

        ' three different methods of generating the list of PCs
        Select Case tcListMethod.SelectedTab.Name
            Case "tpSingle"
                If tbSingle.Text <> "" Then
                    ReDim PC_List(0)
                    PC_List(0) = tbSingle.Text
                End If
            Case "tpAdHoc"
                Dim sc As New StringCollection
                For Each pc As String In tbAdHoc.Lines
                    If pc <> "" Then
                        sc.Add(pc)
                    End If
                Next
                If sc.Count > 0 Then
                    ReDim PC_List(sc.Count - 1)
                    sc.CopyTo(PC_List, 0)
                End If
            Case "tpOU"
                Try
                    Dim net As New GetNetList2
                    PC_List = net.GetPC_Only(tbDomain.Text, tbOU.Text, tbName.Text, cbExcludeServers.Checked)
                Catch ex As Exception
                    MsgBox("Error getting list of PCs" & vbCr & ex.Message, MsgBoxStyle.Exclamation, "Error")
                    If Not IsNothing(tw) Then
                        tw.Close()
                        fs.Close()
                        File.Delete(tbLogfile.Text)
                    End If
                    Exit Sub
                End Try
        End Select

        If IsNothing(PC_List) OrElse PC_List.Length = 0 Then
            MsgBox("Can't generate the list of PCs", MsgBoxStyle.Exclamation, "Error")
            If Not IsNothing(tw) Then
                tw.Close()
                fs.Close()
                File.Delete(tbLogfile.Text)
            End If
            Exit Sub
        End If

        pbStarted.Value = 0
        pbCompleted.Value = 0
        pbStarted.Maximum = PC_List.Length
        pbCompleted.Maximum = PC_List.Length
        ListCount = PC_List.Length

        ' Yeah I know... the ProgressBar is not thread safe, but too bad!
        Control.CheckForIllegalCrossThreadCalls = False

        ' save the existing thread count, and use the value specified
        ThreadPool.GetMaxThreads(maxThreads, portThreads)
        ThreadPool.SetMaxThreads(CInt(nudPoolSize.Value), portThreads)

        ' start the clock
        tbElapse.Text = ""
        tbRemaining.Text = ""
        sWatch.Start()
        Timer1.Enabled = True

        bnStart.Enabled = False
        bnViewLog.Enabled = False
        bnClose.Text = "Cancel"

        ' start 'er up!
        SuccessCount = 0
        TotalCount = 0
        tbStatus.Text = "Started polling " & ListCount & " PCs"

        ' we launch this in a background thread to ensure the form doesn't
        ' appear frozen.  This adds a bit of complexity (a threadpool inside
        ' a background worker), but it makes for a very nice user experience
        BackgroundWorker1.RunWorkerAsync(PC_List)
    End Sub

    '
    ' Just a convenient way to package up all of the data that we need to
    ' send to the delegate
    '
    Public Class bgPackage
        Public waitHandle As AutoResetEvent
        Public RemotePC As String
    End Class

    '
    ' Manages the work queue.  It puts the initial items into the queue and
    ' then waits for them to complete
    '
    Private Sub BackgroundWorker1_DoWork(ByVal sender As Object, ByVal e As System.ComponentModel.DoWorkEventArgs) Handles BackgroundWorker1.DoWork
        Dim PC_Name, PC_List() As String
        Dim p As bgPackage
        Dim i As Integer
        Dim worker As BackgroundWorker
        Dim wh() As AutoResetEvent

        p = Nothing
        PC_List = CType(e.Argument, String())
        worker = CType(sender, BackgroundWorker)
        ReDim wh(PC_List.Length - 1)

        i = 0
        For Each PC_Name In PC_List
            wh(i) = New AutoResetEvent(False)

            ' Setup the "package" that contains the information
            ' we need to send to the delegate
            p = New bgPackage
            p.waitHandle = wh(i)
            p.RemotePC = PC_Name

            ' put it in the queue
            ThreadPool.QueueUserWorkItem(AddressOf StartIt, p)
            i += 1
        Next

        ' The "correct" way to wait for multiple threads to finish is to use
        ' the WaitAll() function.  However, that requires that the calling
        ' process be using the "MTA" threading model.  WinForm apps (under 
        ' VB.Net) are STA.  To make it use MTA you must start the WinForm app
        ' from "Sub Main", rather than a Form1.  But then you loose the cool
        ' new "Window Application Framework" GUI (the WinXP look).  So, we
        ' use this silly approach instead.  Sigh... oh well

        For i = 0 To UBound(wh)
            ' if you asked for it to be canceled 
            If worker.CancellationPending Then
                e.Cancel = True
                If Not IsNothing(tw) Then
                    tw.Flush()
                    tw.WriteLine("Canceled: " & Now())
                    tw.WriteLine("Successfully polled " & SuccessCount & " out of " & TotalCount & " PCs" & vbCr)
                    tw.Close()
                    fs.Close()
                    tw = Nothing
                End If

                ' put everything back...
                bnStart.Enabled = True
                bnViewLog.Enabled = True
                bnClose.Text = "Close"

                ' This does NOT terminate the existing threads that are
                ' already running.
                Exit Sub
            End If
            ' wait until each thread has signaled that it's done
            ' (or until an hour has passed)
            wh(i).WaitOne(60 * 60 * 1000, False)
        Next

        ' put the thread pool count back
        ThreadPool.SetMaxThreads(maxThreads, portThreads)

        sWatch.Stop()
        Timer1.Enabled = False
        If Not IsNothing(tw) Then
            tw.Flush()
            tw.WriteLine("Finished: " & Now())
            tw.WriteLine("Successfully polled " & SuccessCount & " out of " & TotalCount & " PCs")
            tw.WriteLine("")
            tw.Close()
            fs.Close()
            tw = Nothing
        End If

        bnStart.Enabled = True
        bnViewLog.Enabled = True
        bnClose.Text = "Close"
        tbStatus.Text = "Poll Complete"
        tbRemaining.Text = ""
        MsgBox("Successfully polled " & SuccessCount & " out of " & TotalCount & " PCs")
    End Sub

    '
    ' Start the polling in such a way that allows us to terminate the
    ' thread if it runs too long. Has the signature of a 
    ' System.Threading.WaitCallback delegate (which is why we had to bundle
    ' all the data it needs into a "package" class)
    '
    Sub StartIt(ByVal State As Object)
        Dim pkg As bgPackage
        Dim wait As Integer

        pkg = DirectCast(State, bgPackage)

        Dim t As New Thread(AddressOf Doit)
        t.Start(State)

        ' not smart, but it is supported
        If PollTimeOut = 0 Then
            wait = -1
        Else
            wait = PollTimeOut * 60 * 1000
        End If

        ' Wait until the thread ends.  But if it takes too long, abort it
        If t.Join(wait) = False Then
            t.Abort()
            ' give it a minute to close down...
            t.Join(60 * 1000)
        End If

        ' signal that we're done
        pkg.waitHandle.Set()
    End Sub

    ' 
    ' This is where the real work get's done.
    '
    Private Sub Doit(ByVal state As Object)
        Dim RemotePC, ProtoType As String
        Dim pkg As bgPackage
        Dim p As New Ping

        pkg = DirectCast(state, bgPackage)
        RemotePC = pkg.RemotePC
        Try
            pbStarted.Increment(1)

            If PingValidation = True Then
                Try
                    If p.Ping(RemotePC, 1000) > 1000 Then
                        If Not IsNothing(tw) And PollLogLevel = PollLogLevels.Full Then tw.WriteLine(Now() & " " & RemotePC & ": Didn't answer a ping")
                        ' this is not an "error", it's an expected condition
                        Exit Sub
                    End If
                Catch ex As Exception
                    ' Ping might fail on bad DNS lookup
                    If Not IsNothing(tw) And PollLogLevel >= PollLogLevels.Errors Then
                        If PollLogLevel = PollLogLevels.Full Then tw.Write(Now() & " ")
                        tw.WriteLine(RemotePC & ": " & ex.Message.TrimEnd(vbCrLf.ToCharArray))
                    End If
                    Exit Sub
                End Try
            End If

            Try
                ' do it! (copy the files)
                For Each fName As String In Directory.GetFiles(Path.GetDirectoryName(tbSource.Text), Path.GetFileName(tbSource.Text), SearchOption.TopDirectoryOnly)
                    ProtoType = ConvUNC(RemotePC, tbDestination.Text & Path.GetFileName(fName))
                    If File.Exists(ProtoType) Then
                        File.Delete(ProtoType)
                    End If
                    File.Copy(fName, ProtoType)
                Next

                If Not IsNothing(tw) And PollLogLevel = PollLogLevels.Full Then tw.WriteLine(Now & " " & RemotePC & ": OK")
                SuccessCount = Interlocked.Increment(SuccessCount)

            Catch ex As Exception
                If Not IsNothing(tw) And PollLogLevel >= PollLogLevels.Errors Then
                    If PollLogLevel = PollLogLevels.Full Then tw.Write(Now() & " ")
                    If IsNothing(ex.InnerException) Then
                        tw.WriteLine(RemotePC & ": Error - " & ex.Message.TrimEnd(vbCrLf.ToCharArray))
                    Else
                        tw.WriteLine(RemotePC & ": Error - " & ex.Message.TrimEnd(vbCrLf.ToCharArray) & ", " & ex.InnerException.Message.TrimEnd(vbCrLf.ToCharArray))
                    End If
                End If
                Exit Sub
            End Try
        Catch ex_abort As ThreadAbortException
            ' did we have to abort this thread for taking too long?
            If Not IsNothing(tw) And PollLogLevel >= PollLogLevels.Errors Then
                If PollLogLevel = PollLogLevels.Full Then tw.Write(Now() & " ")
                tw.WriteLine(RemotePC & ": Exceeded PollTimeOut limit")
            End If
        Catch ex As Exception
            If Not IsNothing(tw) And PollLogLevel >= PollLogLevels.Errors Then
                If PollLogLevel = PollLogLevels.Full Then tw.Write(Now() & " ")
                If IsNothing(ex.InnerException) Then
                    tw.WriteLine(RemotePC & ": Internal Error - " & ex.Message.TrimEnd(vbCrLf.ToCharArray))
                Else
                    tw.WriteLine(RemotePC & ": Internal Error - " & ex.Message.TrimEnd(vbCrLf.ToCharArray) & ", " & ex.InnerException.Message.TrimEnd(vbCrLf.ToCharArray))
                End If
            End If
        Finally
            pbCompleted.Increment(1)
            TotalCount = Interlocked.Increment(TotalCount)
        End Try
    End Sub

    '
    ' View the results of the polling log using NotePad
    '
    Private Sub bnViewLog_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles bnViewLog.Click
        If [File].Exists(tbLogfile.Text) Then
            Shell("notepad.exe " & tbLogfile.Text, AppWinStyle.NormalFocus)
        Else
            MsgBox("Could not find logfile", MsgBoxStyle.Exclamation, "Warning")
        End If
    End Sub

    '
    ' Browse for the location of the Poll Log file
    '
    Private Sub bnLogFile_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles bnLogFile.Click
        sfdLogFile.AddExtension = True
        sfdLogFile.FileName = tbLogfile.Text
        sfdLogFile.DefaultExt = "txt"
        sfdLogFile.Filter = "Text files (*.txt)|*.txt|All files (*.*)|*.*"
        sfdLogFile.InitialDirectory = My.Settings.LogDir
        sfdLogFile.CheckFileExists = False
        If sfdLogFile.ShowDialog() = Windows.Forms.DialogResult.OK Then
            tbLogfile.Text = sfdLogFile.FileName
        End If
    End Sub

    '
    ' Keeps the elapse time running in the status boxes
    '
    Private Sub Timer1_Tick(ByVal sender As Object, ByVal e As System.EventArgs) Handles Timer1.Tick
        Static completed As Integer = 0

        If sWatch.IsRunning Then
            Dim Current As TimeSpan

            Current = sWatch.Elapsed
            If Current.Hours > 0 Then
                tbElapse.Text = [String].Format("{0:0} hr{1}, {2:0} min, {3:0} sec", Current.Hours, IIf(Current.Hours > 1, "s", ""), Current.Minutes, Current.Seconds)
            Else
                tbElapse.Text = [String].Format("{0:0} min, {1:0} sec", Current.Minutes, Current.Seconds)
            End If

            ' Wait until we've successfully completed at least 10 scans (or
            ' 10% of the total scans) before we attempt to estimate the
            ' remaining time
            If TotalCount >= 10 Or TotalCount / ListCount > 0.1 Then
                Dim Estimated As TimeSpan

                ' only update this when a completion has occurred
                If completed <> TotalCount Then
                    Estimated = TimeSpan.FromSeconds((Current.TotalSeconds * ListCount) / TotalCount)
                    Estimated = Estimated.Subtract(Current)
                    If Estimated.TotalSeconds > 0 Then
                        If Estimated.Hours > 0 Then
                            tbRemaining.Text = [String].Format("{0:0} hr{1}, {2:0} min, {3:0} sec", Estimated.Hours, IIf(Estimated.Hours > 1, "s", ""), Estimated.Minutes, Estimated.Seconds)
                        Else
                            tbRemaining.Text = [String].Format("{0:0} min, {1:0} sec", Estimated.Minutes, Estimated.Seconds)
                        End If
                    End If
                End If
            End If

            If completed <> TotalCount Then
                tbStatus.Text = [String].Format("Completed {0} (out of {1})", TotalCount, ListCount)
                completed = TotalCount
            End If
        End If
    End Sub

    Private Sub bnBrowseSource_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles bnBrowseSource.Click
        ofdSource.DereferenceLinks = False
        ofdSource.Multiselect = False
        ofdSource.Filter = "All files (*.*)|*.*"
        If ofdSource.ShowDialog() <> Windows.Forms.DialogResult.OK Then
            Exit Sub
        End If
        tbSource.Text = ofdSource.FileName
    End Sub

    Private Sub bnBrowseDest_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles bnBrowseDest.Click
        If fbdDestination.ShowDialog() <> Windows.Forms.DialogResult.OK Then
            Exit Sub
        End If
        tbDestination.Text = fbdDestination.SelectedPath
    End Sub

    ' Convert a "local" filename to a UNC remote filename
    Protected Function ConvUNC(ByVal RemotePC As String, ByVal Filename As String) As String
        Dim ans As String = ""

        If RemotePC = "" Then
            Return Filename
        End If

        ' drive letter
        If Filename.Length = 2 AndAlso Filename.EndsWith(":") Then
            Return "\\" & RemotePC & "\" & Filename.Substring(0, 1) & "$"
        End If

        If Filename.Length >= 3 Then
            Return "\\" & RemotePC & "\" & Filename.Substring(0, 1) & "$" & Filename.Substring(2)
        End If

        Return Filename
    End Function
End Class
